/*
 *
 * hal_interrupt_cause.c
 * Description:
 * Implementation functions that retrieve the interrupt cause for DOCSIS Interrupts
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#define _HAL_INTERRUPT_CAUSE_C_

/*! \file hal_interrupt_handlers_u.c
    \brief Implementation functions that retrieve the interrupt cause for DOCSIS
    \Interrupts. 
    \IMPORTANT:
    \   A. This code will run as part of a loadable kernel module, and should 
    \      be used in kernel-space only.
    \   B. These functions assume that they are protected externally, meaning 
    \      that their code is executed in "atomic" form.
*/

/**************************************************************************/
/*      INCLUDES:                                                         */
/**************************************************************************/
#include <linux/module.h>	/* module */
#include <linux/interrupt.h>
#include <linux/irq.h>
#include "hal_interrupt_cause.h"
#include "hal_mac_interrupt_regs.h"
#include "hal_phy_basic_regs.h"
#include "hal_phy_interrupt_regs.h"
#include <asm/arch/puma5/puma5_hardware.h>

/*#define DEBUG_CAUSE_PRINTS*/

/**************************************************************************/
/*      EXTERNS Declaration:                                              */
/**************************************************************************/
extern unsigned int pgaGrtIntNum;

/**************************************************************************/
/*      DEFINES:                                                          */
/**************************************************************************/

#define PHY_NUMBER_OF_RECIEVERS                     4
#define PHY_NUMBER_OF_TX_INTERRUPTS                 5

#define PUMA5_HW_REV_1_X   0
#define PUMA5_HW_REV_2_0   1

#define FEC_LOCK_VAL                                1
#define FEC_LOST_VAL                                0
#define NUM_OF_DS                                   4

#define PHY_RX1                                     0
#define PHY_RX2                                     1
#define PHY_RX3                                     2
#define PHY_RX4                                     3

typedef struct{
    int currState;
    int prevState;
}PhyFecRegsStatus_t;

typedef struct{
    Bool isFecLostMasked[NUM_OF_DS];
    Bool isFecLockMasked[NUM_OF_DS];
}PhyFecLockLostStatus_t;

    PhyFecRegsStatus_t PhyFecRegsStatusArr [NUM_OF_DS];
    PhyFecLockLostStatus_t PhyFecLockLost;

/**************************************************************************/
/*      LOCAL DECLARATIONS:                                               */
/**************************************************************************/
static Int32 HAL_IsrMaskPhyInterrupt(Uint32 cause, Bool maskOp);

static Int32 HAL_GetPhyReg16K(Uint32 phyRegAddr, Uint16 *pValue);
static Int32 HAL_SetPhyReg16K(Uint32 phyRegAddr, Uint16 value);


/**************************************************************************/
/*      LOCAL VARIABLES:                                                  */
/**************************************************************************/


/**************************************************************************/
/*      INTERFACE FUNCTIONS Implementation:                               */
/**************************************************************************/


/**********************************/               
/*  MAC Errors Interrupt Group    */               
/**********************************/               

/**************************************************************************/
/*! \fn Int32 HAL_IsrGetCauseMacErrorGroup(IntCauseErrorGroup_t *pCause)
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC-error group interrupt-line.
 *  \param[out] pCause pointer to the structure that will hold the interrupt cause(s).
 *  \return OK if successful, reject or fail otherwise                                     
 **************************************************************************/
Int32 HAL_IsrGetCauseMacErrorGroup(IntCauseErrorGroup_t *pCause)
{
    *(volatile Uint32*)(pCause) = *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                            MAC_ERROR_INTERRUPT_1_STATUS_REG);

    *(((volatile Uint32*)pCause) + 1) = *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                            MAC_ERROR_INTERRUPT_2_STATUS_REG);

    *(((volatile Uint32*)pCause) + 2) = *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                            MAC_ERROR_INTERRUPT_3_STATUS_REG);

    *(((volatile Uint32*)pCause) + 3) = *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                            MAC_ERROR_INTERRUPT_4_STATUS_REG);

    return 0;
}


/****************************************/               
/*    MAC UCD-Change Interrupt group    */               
/****************************************/               

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetCauseUcdChange()
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC UCD Change interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 **************************************************************************/
Uint32 HAL_IsrGetCauseUcdChange()
{
    return *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                  MAC_UCD_CHANGE_INTERRUPT_STATUS_REG);
}

/***********************************/               
/*    MAC US FW Interrupt group    */               
/***********************************/               

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetCauseUsFw()
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC US FW interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 **************************************************************************/
Uint32 HAL_IsrGetCauseUsFw()
{
    return *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                  MAC_US_FW_INTERRUPT_STATUS_REG);
}

/***********************************/               
/*    MAC DS FW Interrupt group    */               
/***********************************/               

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetCauseDsFw()
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC DS FW interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 **************************************************************************/
Uint32 HAL_IsrGetCauseDsFw()
{
    return *(volatile Uint32*)(MAC_KERNEL_BASE_ADDRESS + 
                                  MAC_DS_FW_INTERRUPT_STATUS_REG);
}



/***********************************/
/*         PHY  Interrupts         */
/***********************************/

/**************************************************************************/
/*! \fn Int32 HAL_IsrResetPhyFecDB()
 **************************************************************************
 *  \brief Reset Phy-Rx database.
 *  \param[in] none.
 *  \return OK if successful, reject or fail otherwise.
 **************************************************************************/
Int32 HAL_IsrResetPhyFecDB()
{
    memset((void *)&PhyFecRegsStatusArr, 0, (sizeof(PhyFecRegsStatus_t)*NUM_OF_DS));
    memset((void *)&PhyFecLockLost, 1, sizeof(PhyFecLockLostStatus_t));
    return 0;
}

/**************************************************************************/
/*! \fn Uint32 HAL_IsrGetCausePhy()
 **************************************************************************
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \PHY interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 **************************************************************************/
Uint32 HAL_IsrGetCausePhy()
{
    Uint32 pCause;
    Uint16 valRx1Rx2, valRx3Rx4;
    Uint16 valFec [NUM_OF_DS];
    Uint32 bitVal;
    int i;

    pCause = 0;

#ifdef DEBUG_CAUSE_PRINTS
    printk(KERN_INFO "--> %s: \n", __FUNCTION__);
#endif
    /*check FEC masking; Registers: 0x0c2, 0x0c4*/
    HAL_GetPhyReg16K(PHY_RX1_RX2_INTERRUPT_ENABLE, &valRx1Rx2);
    HAL_GetPhyReg16K(PHY_RX3_RX4_INTERRUPT_ENABLE, &valRx3Rx4);
    /*check the relevant registers for FEC loss/lock; Registers: 0x0a3, 0x1a3, 0x2a3, 0x3a3*/
    for(i=0; i<NUM_OF_DS; i++)
    {
        HAL_GetPhyReg16K(PHY_RX_FEC_STATUS(i+1), &valFec[i]);
    }

    /*updating database*/
    for(i=0; i<NUM_OF_DS; i++)
    {
        Uint16 annexVal;

        bitVal = (valFec[i]&(1<<(PHY_RX_FEC_STATUS_TR_SYNC)));
        bitVal = (bitVal>>PHY_RX_FEC_STATUS_TR_SYNC);

        /* Euro patch: FEC A Status Register is inverted to FEC B Status Register */
        HAL_GetPhyReg16K(PHY_RX_CH_REG_OFFSET(i, PHY_QAM_MODE_0_REGISTER), &annexVal);
        annexVal = annexVal & (1<<PHY_QAM_ANNEX_TYPE);
        if(annexVal)/*0 - Annex B, 1 - Annex A*/
        {
            if (bitVal)
            {
                bitVal = FEC_LOST_VAL;
            }
            else
            {
                bitVal = FEC_LOCK_VAL;
            }
        }

        if (bitVal != PhyFecRegsStatusArr[i].currState)
        {
            Uint32 result;
#ifdef DEBUG_CAUSE_PRINTS
            printk( "-->%s: i=%d: bitVal = %d, currState  = %d \n",__FUNCTION__, i, bitVal, PhyFecRegsStatusArr[i].currState);
#endif

            /*updating database*/
            PhyFecRegsStatusArr[i].prevState = PhyFecRegsStatusArr[i].currState;
            PhyFecRegsStatusArr[i].currState = bitVal;

            if(bitVal == FEC_LOST_VAL)/*0*/
            {
#ifdef DEBUG_CAUSE_PRINTS
                printk( "-->%s: bitVal = FEC_LOST_VAL\n", __FUNCTION__);
#endif

                switch(i)
                {
                    case PHY_RX1:
                        result = (valRx1Rx2 & (1<<PHY_RX1_FEC_LOST));
                        if(result | !(PhyFecLockLost.isFecLostMasked[0]))
                            pCause |= (1<<PHY_RX1_INT_FEC_LOST);
                        break;
                    case PHY_RX2:
                        result = (valRx1Rx2 & (1<<PHY_RX2_FEC_LOST));
                        if(result | !(PhyFecLockLost.isFecLostMasked[1]))
                            pCause |= (1<<PHY_RX2_INT_FEC_LOST);
                        break;
                    case PHY_RX3:
                        result = (valRx3Rx4 & (1<<PHY_RX3_FEC_LOST));
                        if(result | !(PhyFecLockLost.isFecLostMasked[2]))
                            pCause |= (1<<PHY_RX3_INT_FEC_LOST);
                        break;
                    case PHY_RX4:
                        result = (valRx3Rx4 & (1<<PHY_RX4_FEC_LOST));
                        if(result | !(PhyFecLockLost.isFecLostMasked[3]))
                            pCause |= (1<<PHY_RX4_INT_FEC_LOST);
                        break;
                    default:/*add print*/
                        break;
                }
            }

            if(bitVal == FEC_LOCK_VAL)/*1*/
            {
#ifdef DEBUG_CAUSE_PRINTS
                printk( "-->%s: bitVal = FEC_LOCK_VAL\n",__FUNCTION__);
#endif

                switch(i)
                {
                    case PHY_RX1:
                        result = (valRx1Rx2 & (1<<PHY_RX1_FEC_LOCKED));
                        if(result | !(PhyFecLockLost.isFecLockMasked[0]))
                            pCause |= (1<<PHY_RX1_INT_FEC_LOCK);
                        break;
                    case PHY_RX2:
                        result = (valRx1Rx2 & (1<<PHY_RX2_FEC_LOCKED));
                        if(result | !(PhyFecLockLost.isFecLockMasked[1]))
                            pCause |= (1<<PHY_RX2_INT_FEC_LOCK);
                        break;
                    case PHY_RX3:
                        result = (valRx3Rx4 & (1<<PHY_RX3_FEC_LOCKED));
                        if(result | !(PhyFecLockLost.isFecLockMasked[2]))
                            pCause |= (1<<PHY_RX3_INT_FEC_LOCK);
                        break;
                    case PHY_RX4:
                        result = (valRx3Rx4 & (1<<PHY_RX4_FEC_LOCKED));
                        if(result | !(PhyFecLockLost.isFecLockMasked[3]))
                            pCause |= (1<<PHY_RX4_INT_FEC_LOCK);
                        break;
                    default:/*add print*/
                        break;
                }
            }

        }/*if*/
    }/*for*/

#ifdef DEBUG_CAUSE_PRINTS
    printk( "-->%s: pCause = 0x%x \n",__FUNCTION__, pCause);
#endif

    /*returning bitmask*/
    return pCause;

}

/**************************************************************************/
/*! \fn Int32 HAL_IsrMaskInterrupt(Uint32 cause, Bool maskOp)
 **************************************************************************
 *  \brief Masks or unmasks a speific interrupt.
 *  \param[in] cause defines the specific interrupt that needs to be 
 *  \                masked/unmasked.
 *  \param[in] maskOp defines whether to mask or unmask. 
 *  \                 True = mask, False= unmask. 
 *  \return OK if successful, reject or fail otherwise                     
 **************************************************************************/
Int32 HAL_IsrMaskInterrupt(Uint32 cause, Bool maskOp)
{
    Uint32 address = MAC_KERNEL_BASE_ADDRESS;
    Uint32 group = HAL_GET_INTERRUPT_GROUP(cause);
    Uint8  bit = HAL_GET_INTERRUPT_BIT(cause);

    switch (group) {
	case MAC_ERR1_INT_GROUP:
		if (bit > MAC_ERR1_INT_FLTR_UCD_US4)
			return -1;
		address += MAC_ERROR_INTERRUPT_1_MASK_REG;
		break;

	case MAC_ERR2_INT_GROUP:
		address += MAC_ERROR_INTERRUPT_2_MASK_REG;
		break;

	case MAC_ERR3_INT_GROUP:
		if (bit > MAC_ERR3_INT_MP_ORD_FIFO_FULL)
			return -1;
		address += MAC_ERROR_INTERRUPT_3_MASK_REG;
		break;

	case MAC_ERR4_INT_GROUP:
		address += MAC_ERROR_INTERRUPT_4_MASK_REG;
		break;

	case MAC_UCD_INT_GROUP:
		if (bit > MAC_UCD_INT_NEW_UCD_US4)
			return -1;
		address += MAC_UCD_CHANGE_INTERRUPT_MASK_REG;
		break;

	case MAC_US_FW_INT_GROUP:
		address += MAC_US_FW_INTERRUPT_MASK_REG;
		break;

	case MAC_DS_FW_INT_GROUP:
		address += MAC_DS_FW_INTERRUPT_MASK_REG;
		break;

	case PHY_INT_GROUP:
			return HAL_IsrMaskPhyInterrupt(cause, maskOp);

    case PHY_PGA_GRT_INT_GROUP:
        if(maskOp == True)
        {
            printk(KERN_INFO "HAL_IsrMaskInterrupt: calling disable_irq(PHY_PGA_GRT_INT_NUM)\n");
            disable_irq(PHY_PGA_GRT_INT_NUM);
        }
        else
        {
            printk(KERN_INFO "HAL_IsrMaskInterrupt: calling enable_irq(PHY_PGA_GRT_INT_NUM)\n");
            enable_irq(PHY_PGA_GRT_INT_NUM);
        }
        return 0;


	default:
		return -1;
	}

    if(maskOp)
        /* mask the interrupt*/
        *((volatile Uint32 *)address) |= (1<<bit);
    else
        /* unmask the interrupt*/
        *((volatile Uint32 *)address) &= ~(1<<bit);
    return 0;
}


/**************************************************************************/
/*      LOCAL FUNCTIONS:                                                  */
/**************************************************************************/
/**************************************************************************/
/*! \fn static Int32 HAL_IsrMaskPhyInterrupt(Uint32 cause, Bool maskOp)
 **************************************************************************
 *  \brief Perform mask or unmask operation.
 *  \param[in] Uint32 cause - interrupts bitmask.
 *  \param[in] Bool maskOp - Mask operation (maskOp = 1) or Unmask operation
 *   (maskOp = 0).
 *  \return OK if successful, reject or fail otherwise.
 **************************************************************************/
static Int32 HAL_IsrMaskPhyInterrupt(Uint32 cause, Bool maskOp)
{
    Uint32 address;
    Uint8  intrBit;
    Uint16 regVal;
    Uint32 bitVal;
    Uint16 valFec [NUM_OF_DS];
    Int32 i;
    Uint32 arrIndex;
    Bool isFecLock;
    Uint32 enCounter, maskCounter;
    Uint32 enCounterPrev, maskCounterPrev;
    Uint32 addressMac;
    Uint32 groupMac;

#ifdef DEBUG_CAUSE_PRINTS
    printk( "-->%s:\n", __FUNCTION__);
#endif

    enCounter       = 0;
    maskCounter     = 0;
    enCounterPrev   = 0;
    maskCounterPrev = 0;
    isFecLock       = False;


    intrBit   = HAL_GET_INTERRUPT_BIT(cause);
    arrIndex  = HAL_PHY_GET_FEC_STATUS_DB_INDX(intrBit);
    isFecLock = HAL_PHY_IS_FEC_LOCK(intrBit);

    addressMac = (MAC_KERNEL_BASE_ADDRESS + MAC_US_FW_INTERRUPT_MASK_REG);
    groupMac   = MAC_US_FW_INT_GROUP;

    /*saving previous state of EN lock/lost DB status*/
    for(i=0; i<NUM_OF_DS; i++)
    {
        if(PhyFecLockLost.isFecLockMasked[i] == False)
            enCounterPrev++;
        else
            maskCounterPrev++;

        if(PhyFecLockLost.isFecLostMasked[i] == False)
            enCounterPrev++;
        else
            maskCounterPrev++;
    }

    switch(intrBit)
    {
    case PHY_RX1_INT_FEC_LOST:
        address = PHY_RX1_RX2_INTERRUPT_ENABLE;
        intrBit = PHY_RX1_FEC_LOST;
        PhyFecLockLost.isFecLostMasked[arrIndex] = maskOp;
#ifdef DEBUG_CAUSE_PRINTS
        printk( "-->%s: intrBit = PHY_RX1_INT_FEC_LOST\n", __FUNCTION__);
#endif
        break;

    case PHY_RX1_INT_FEC_LOCK:
        address = PHY_RX1_RX2_INTERRUPT_ENABLE;
        intrBit = PHY_RX1_FEC_LOCKED;
        PhyFecLockLost.isFecLockMasked[arrIndex] = maskOp;
#ifdef DEBUG_CAUSE_PRINTS
        printk( "-->%s: intrBit = PHY_RX1_FEC_LOCKED\n", __FUNCTION__);
#endif
        break;

    case PHY_RX2_INT_FEC_LOST:
        address = PHY_RX1_RX2_INTERRUPT_ENABLE;
        intrBit = PHY_RX2_FEC_LOST;
        PhyFecLockLost.isFecLostMasked[arrIndex] = maskOp;
#ifdef DEBUG_CAUSE_PRINTS
        printk( "-->%s: intrBit = PHY_RX2_FEC_LOST\n", __FUNCTION__);
#endif
        break;

    case PHY_RX2_INT_FEC_LOCK:
        address = PHY_RX1_RX2_INTERRUPT_ENABLE;
        intrBit = PHY_RX2_FEC_LOCKED;
        PhyFecLockLost.isFecLockMasked[arrIndex] = maskOp;
#ifdef DEBUG_CAUSE_PRINTS
        printk( "-->%s: intrBit = PHY_RX2_FEC_LOCKED\n", __FUNCTION__);
#endif
        break;

    case PHY_RX3_INT_FEC_LOST:
        address = PHY_RX3_RX4_INTERRUPT_ENABLE;
        intrBit = PHY_RX3_FEC_LOST;
        PhyFecLockLost.isFecLostMasked[arrIndex] = maskOp;
#ifdef DEBUG_CAUSE_PRINTS
        printk( "-->%s: intrBit = PHY_RX3_FEC_LOST\n", __FUNCTION__);
#endif
        break;

    case PHY_RX3_INT_FEC_LOCK:
        address = PHY_RX3_RX4_INTERRUPT_ENABLE;
        intrBit = PHY_RX3_FEC_LOCKED;
        PhyFecLockLost.isFecLockMasked[arrIndex] = maskOp;
#ifdef DEBUG_CAUSE_PRINTS
        printk( "-->%s: intrBit = PHY_RX3_FEC_LOCKED\n", __FUNCTION__);
#endif
        break;

    case PHY_RX4_INT_FEC_LOST:
        address = PHY_RX3_RX4_INTERRUPT_ENABLE;
        intrBit = PHY_RX4_FEC_LOST;
        PhyFecLockLost.isFecLostMasked[arrIndex] = maskOp;
#ifdef DEBUG_CAUSE_PRINTS
        printk( "-->%s: intrBit = PHY_RX4_FEC_LOST\n", __FUNCTION__);
#endif
        break;

    case PHY_RX4_INT_FEC_LOCK:
        address = PHY_RX3_RX4_INTERRUPT_ENABLE;
        intrBit = PHY_RX4_FEC_LOCKED;
        PhyFecLockLost.isFecLockMasked[arrIndex] = maskOp;
#ifdef DEBUG_CAUSE_PRINTS
        printk( "-->%s: intrBit = PHY_RX4_FEC_LOCKED\n", __FUNCTION__);
#endif
        break;

    default:
        return -1;
    }

#ifdef DEBUG_CAUSE_PRINTS
printk( "-->%s: maskOp = %d, isFecLock = %d\n", __FUNCTION__, maskOp, isFecLock);
#endif

    /* Checking the new and updated counters...*/
    for(i=0; i<NUM_OF_DS; i++)
    {
        if(PhyFecLockLost.isFecLostMasked[i] == False)
            enCounter++;
        else
            maskCounter++;

        if(PhyFecLockLost.isFecLockMasked[i] == False)
            enCounter++;
        else
            maskCounter++;
    }

    if((maskCounter == NUM_OF_DS*2) && (maskCounterPrev == NUM_OF_DS*2-1))
    {/* set bit 30 to disable */
        *((volatile Uint32 *)addressMac) |= (1<<MAC_US_FW_INT_PHY_FEC_LOST_LOCK);
#ifdef DEBUG_CAUSE_PRINTS
printk( "-->%s: Disable FW interrupt\n", __FUNCTION__);
#endif
    }

    if((enCounter == 1) && (enCounterPrev == 0))
    {   /*set bit 30 to enable*/
        *((volatile Uint32 *)addressMac) &= ~(1<<MAC_US_FW_INT_PHY_FEC_LOST_LOCK);
#ifdef DEBUG_CAUSE_PRINTS
		printk( "-->%s: Enable FW interrupt\n", __FUNCTION__);
#endif
    }


    /* Note: in the PHY core, the logic of Mask/Unmask is:  */
    /*       intrBit=1 ==> Interrupt Unmasked (enabled)     */
    /*       intrBit=0 ==> Interrupt Masked (disabled)      */
	if (system_rev != PUMA5_HW_REV_2_0) /* this register was changed in rev2*/
	{
		HAL_GetPhyReg16K(address,&regVal);
		if (maskOp) /* mask (disable) the interrupt */
			regVal &= (~(1<<intrBit));
		else        /* unmask (enable) the interrupt*/
			regVal |= (1<<intrBit);
		HAL_SetPhyReg16K(address, regVal);
	}

    /*check the relevant registers for FEC loss/lock; Registers: 0x0a3, 0x1a3, 0x2a3, 0x3a3*/
    for(i=0; i<NUM_OF_DS; i++)
    {
        HAL_GetPhyReg16K(PHY_RX_FEC_STATUS(i+1), &valFec[i]);
    }

#ifdef DEBUG_CAUSE_PRINTS
    printk(KERN_INFO "-->%s: updating database...\n", __FUNCTION__);
#endif
    /*updating database*/
    for(i=0; i<NUM_OF_DS; i++)
    {
        Uint16 annexVal;

        bitVal = (valFec[i]&(1<<(PHY_RX_FEC_STATUS_TR_SYNC)));
        bitVal = (bitVal>>PHY_RX_FEC_STATUS_TR_SYNC);

        /* Euro patch: FEC A Status Register is inverted to FEC B Status Register */
        HAL_GetPhyReg16K(PHY_RX_CH_REG_OFFSET(i, PHY_QAM_MODE_0_REGISTER), &annexVal);
        annexVal = annexVal & (1<<PHY_QAM_ANNEX_TYPE);
        if(annexVal)/*0 - Annex B, 1 - Annex A*/
        {
            if (bitVal)
            {
                bitVal = FEC_LOST_VAL;
            }
            else
            {
                bitVal = FEC_LOCK_VAL;
            }
        }

        if (bitVal != PhyFecRegsStatusArr[i].currState)
        {
            /*updating database*/
            PhyFecRegsStatusArr[i].prevState = PhyFecRegsStatusArr[i].currState;
            PhyFecRegsStatusArr[i].currState = bitVal;
        }
    }

    return 0;
}


static Int32 HAL_GetPhyReg16K(Uint32 phyRegAddr, Uint16 *pValue)
{
    Uint32 regOffset;

    regOffset = phyRegAddr*2;/*2-bytes for each reg*/
    if(regOffset >= HAL_PHY_ADDRESS_SPACE_SIZE)
        return -1;

    *pValue = *(volatile Uint16*)(PHY_KERNEL_BASE_ADDRESS + regOffset);

#ifdef DEBUG_CAUSE_PRINTS
    printk(KERN_INFO "%s: phyRegAddr = 0x%x, regOffset = %d, value = (0x%x)\n", __FUNCTION__, phyRegAddr, regOffset, *pValue);
#endif

    return 0;
}

static Int32 HAL_SetPhyReg16K(Uint32 phyRegAddr, Uint16 value)
{
    Uint32 regOffset;

    regOffset = phyRegAddr*2;/*2-bytes for each reg*/
    if(regOffset >= HAL_PHY_ADDRESS_SPACE_SIZE)
        return -1;

    *(volatile Uint16*)(PHY_KERNEL_BASE_ADDRESS + regOffset) = value;

#ifdef DEBUG_CAUSE_PRINTS
    printk(KERN_INFO "%s: phyRegAddr =  0x%x, regOffset = %d, value = (0x%x)\n", __FUNCTION__, phyRegAddr, regOffset, value);
#endif

    return 0;
}

